
#include "meme.h"
#include "memetype.h"
#include <iostream>
#include <cstdio>

using namespace std;

MemeScope::MemeScope(MemeScope *parent)
	: parent_ ( parent ), hasReturned_( false ), hasBreak_( false )
{
}

MemeScope::~MemeScope()
{
	// map and gc handle it all
}

void MemeScope::dump(void) // dump to stdout for debugging
{
	std::map<std::string,Value>::iterator it;
	for( it = symbols_.begin(); it != symbols_.end(); it++ )
		cout << it->first << "=" << it->second.toString() << endl;
}

// define a new variable, will replace anything at that name in the scope
MemeRef MemeScope::define(const char *name, const Value& val)
{
	Value *actual = new Value( val ); // NOTE: need copy construct that doesn't mess with refcount
	MemeRef theRef( actual );
	bind( name, theRef );
	return theRef;
}

// bind a ref to a variable
void MemeScope::bind(const char *name, MemeRef theRef)
{
	std::map<std::string,Value>::iterator it = symbols_.find( name );
	if( it == symbols_.end() )
		symbols_.insert( SymEntry(name, Value(valueRef, theRef) ) );
	else
		it->second = Value(valueRef, theRef);
}

Value MemeScope::assign(const char *name, const Value& val)
{
	//  call error if try to assign to keyword
	std::map<std::string,Value>::iterator it = symbols_.find( name );
	if( it == symbols_.end() ) {
		error("undefined identifier %s",name);
		// allocate a new one to be ref counted and put ref in map
		return Value();
	} else {
		it->second.getPrimitive()->lassign( it->second, val );
		return it->second;
	}
}


Value* MemeScope::lookup(const char *name)
{
	std::map<std::string,Value>::iterator it = symbols_.find( name );
	if( it == symbols_.end() )
	{
		if( parent_ )
			return parent_->lookup(name);
		else
			return NULL;
	}
	else
		return &(it->second);
}

// we use scope to implement stack frames, and return and
// other types of unwinding.
// this call is used to signal a return.
void MemeScope::func_return(const Value& val){
	// NOTE: good place for an assertion that we have not returned.
	return_ = val;
	hasReturned_ = true;
}
bool MemeScope::unwind( void ) const
{
	return hasReturned_ || hasBreak_;
}

void MemeScope::set_break(void)
{
	hasBreak_ = true;
}
void MemeScope::clr_break(void)
{
	hasBreak_ = false;
}

Value MemeScope::getReturn(void) const {
	return return_;
}

// these are not abstract, they all throw
// errors if called on an object that does
// not implement
MemePrimitive::MemePrimitive()
{

}


MemePrimitive::~MemePrimitive()
{

}

//  this should have variable arguments
Value MemePrimitive::ini(const Value& x)
{
	return x;
}

Value MemePrimitive::add(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::sub(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::neg(const Value& x)
{
	error("unsupported message");
	return Value();
}



Value MemePrimitive::mul(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::div(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::mod(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}



Value MemePrimitive::cmp(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::len(const Value& a)
{
	error("unsupported message");
	return Value();
}

Value MemePrimitive::idx(const Value& a, const Value& i)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::ins(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}

Value MemePrimitive::cat(const Value& a, const Value& b)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::str(const Value& a)
{
	error("unsupported message");
	return Value();
}


Value MemePrimitive::num(const Value& a)
{
	error("unsupported message");
	return Value();
}

void MemePrimitive::lassign(const Value& left, const Value& right)
{
	error("not an lvalue");
}


MemeString::MemeString(void)
{

}

MemeString::MemeString(const char *str)
	: str_(str)
{
}

MemeString::MemeString(const char *str, int len)
	: str_(str, len)
{
}

MemeString::MemeString(const MemeString& ms)
	: str_(ms.str_)
{

}

MemeString::~MemeString()
{

}

const MemeString& MemeString::operator = (const MemeString& ms)
{
	str_.assign( ms.str_ );
	return *this;
}


const char *MemeString::getText(void) const
{
	return str_.c_str();
}

int MemeString::getLength(void) const
{
	return str_.size();
}

void MemeString::setText( const char *str )
{
	str_.assign(str);
}


void MemeString::append( const char *str )
{
	str_.append( str );
}

// create and store a string and return a reference to it.
MemeRef MemeString::newString( const char *str )
{
	MemeString *ms = new MemeString( str );
	if( !ms ) error("out of memory");
	return MemeRef( ms );
}

MemeRef MemeString::newString( const char *str, int len )
{
	MemeString *ms = new MemeString( str, len );
	if( !ms ) error("out of memory");
	return MemeRef( ms );
}


MemeString* MemeString::derefString( const MemeRef& ref ) {
	return (MemeString *)ref.target();
}

const char* MemeString::strFromRef( const MemeRef& ref ) {
	return derefString( ref )->getText();
}

Value::Value(void)
	: type(valueNull)
{
}

Value::Value(int v)
	: type(valueInt), i(v)
{
}

Value::Value(double v)
	: type(valueReal), r(v)
{
}

Value::Value(bool flag)
	: type(valueBool), b(flag)
{

}

Value::Value(valueEnum t, const MemeRef& ref)
	: type(t), ref_(ref)
{
}

Value::Value(const Value& v )
{
	assign( v );
}



const Value& Value::operator = (const Value& x )
{
	assign(x);
	return *this;
}

int Value::toInt(void) const {
	Value num = getPrimitive()->num( *this );
	if( num.type == valueInt )
		return num.i;
	else if( num.type == valueReal )
		return (int)num.r;
	else // null
		return 0;
}
double Value::toReal(void) const {
	Value num = getPrimitive()->num( *this );
	if( num.type == valueInt )
		return num.i;
	else if( num.type == valueReal )
		return num.r;
	else // null
		return 0;

}
const char *Value::toString(void) const {
	return MemeString::strFromRef( getPrimitive()->str( *this ).ref_ );
}

bool Value::toBool( ) const {
	return toInt() != 0;
}
enum {kMaxPrimitives=512};
static MemePrimitive *sPrimitive[ kMaxPrimitives ];
void Value::registerPrimitive(valueEnum type, MemePrimitive *prim)
{
	sPrimitive[type] = prim;
}

MemePrimitive* Value::getPrimitive(void) const
{
	return sPrimitive[ type ];
}

void Value::assign(const Value& val)
{
	type = val.type;
	if( type == valueInt )
		i = val.i;
	else if( type == valueReal )
		r = val.r;
	else if( type == valueBool )
		b = val.b;
	else if( type == valueStr || type == valueRef || type == valueArray
			|| type == valueHash || type == valueAppender || type == valueFunc )
		ref_ = val.ref_;
	// else null
}

Value Value::value(void) const
{
	if( type == valueRef )
		return *(Value *)ref_.target();
	else
		return *this;
}

void Value::print(void) const {
	Value sv = getPrimitive()->str( * this );
	const char *s = MemeString::strFromRef( sv.ref_ );
	cout << s;
}

void Value::println(void) const {
	print();
	cout << endl;
}

//struct nodeTypeTag *p_;
MemeFunc::MemeFunc( MemeScope *parent, struct nodeTypeTag *p) {
	parent_ = parent;
	p_ = p;
}

MemeFunc::~MemeFunc() {
	// we are not responsible for freeing crap.
}

void MemeFunc::SetupLocals(std::vector<Value> & args, MemeScope & locals)
{
#if MEME_TRACE
	std::cout << "args.size() == " << args.size() << std::endl; // TODO: disable
#endif
    nodeType *param = p_->func.params;
    int nparams = 0;
    while(param){
        if(nparams >= (int)(args.size()))
            error("not enough parameters");

        Value v =  args[nparams].value();
#if MEME_TRACE
        std::cout << ":" << param->param.name << "=" << v.toString() << std::endl; // TODO: disable
#endif
        locals.define(param->param.name,v );
        nparams++;
        param = param->next;
    }
    if(nparams != (int)(args.size()))
        warning("extra parameters");

}


MemeNativeFunc::MemeNativeFunc(MemeScope *parent, struct nodeTypeTag *p)
: MemeFunc(parent,p)
{

}
MemeNativeFunc::~MemeNativeFunc()
{

}

Value MemeNativeFunc::call( std::vector<Value>& args ) {
    MemeScope locals(parent_);
    SetupLocals(args, locals);
    Value rval = ex(locals, p_->func.impl );
    if( locals.unwind() )
    	rval = locals.getReturn();
    return rval;
}

MemeCppFunc::MemeCppFunc(MemeScope *parent, struct nodeTypeTag *p, MemeCppProc proc)
: MemeFunc(parent,p)
{
	proc_ = proc;
}

MemeCppFunc::~MemeCppFunc()
{
}


Value MemeCppFunc::call(std::vector<Value> & args)
{
	MemeScope locals(parent_);
	SetupLocals(args, locals);
	Value result;
	try
	{
		result =  proc_( locals );
	}
	catch(...)
	{
		result = Value();
		warning("builtin threw exception");
	}
	return result;
}






